from itertools import zip_longest
from typing import List, Union

import numpy as np
from qcio import (
    CalcType,
    DualProgramInput,
    Inputs,
    OptimizationResults,
    ProgramArgs,
    ProgramArgsSub,
    ProgramInput,
    ProgramOutput,
    Results,
    SinglePointResults,
    StructuredInputs,
)
from qcop import compute as qcop_compute

from .app import bigchem

__all__ = [
    "compute",
    # "output_to_input",
    "assemble_hessian",
    "frequency_analysis",
]


@bigchem.task
def compute(
    program: Union[str, Inputs],
    inp_obj: Union[Inputs, str],
    **kwargs,
) -> ProgramOutput[Inputs, Results]:
    """Wrapper around qcop.compute.

    Checks first and second argument order as they may be reversed due to chaining.
    For example, output_to_input returns an output object, but compute expects program
    as the first argument. Chains can only pass the output object as the first argument
    to the next task in the chain. This wrapper allows the user to pass the program
    first or second.
    """
    if isinstance(inp_obj, str):
        # If the first argument is a string, then the second argument is the input
        program, inp_obj = inp_obj, program
    return qcop_compute(program, inp_obj, **kwargs)


@bigchem.task
def output_to_input(
    output: ProgramOutput[StructuredInputs, Results],
    calctype: CalcType,
    program_args: Union[ProgramArgs, ProgramArgsSub],
) -> Union[ProgramInput, DualProgramInput]:
    """Propagate output values from a calculation onto a new input object.

    Args:
        output: SinglePointOutput, OptimizationOutput, or ProgramFailure object
        program_args: QCProgramArgs or SubProgramArgs object
        calctype: Calculation type for the new input

    NOTE: This task requires additional work to cover more general cases. This skeleton
        is primarily to give an initial example of basic multi-package geometry
        optimization.
    """
    input_model = (
        ProgramInput if isinstance(program_args, ProgramArgs) else DualProgramInput
    )
    if output.input_data.calctype in {CalcType.optimization, CalcType.transition_state}:
        assert isinstance(output.results, OptimizationResults)  # mypy
        # Take final geometry from optimization and pass to next input
        return input_model(
            molecule=output.results.final_molecule,
            calctype=calctype,
            **program_args.model_dump(),
        )
    else:
        # TODO: Add wavefunction passing for TeraChem somewhere? Not here...
        raise NotImplementedError(
            f"No implementation for transforming {output.__class__.__name__} objects "
            f"into {input_model.__name__} objects yet."
        )


@bigchem.task
def assemble_hessian(
    gradients: List[ProgramOutput[ProgramInput, SinglePointResults]], dh: float
) -> ProgramOutput[ProgramInput, SinglePointResults]:
    """Assemble hessian from an array of gradient computations

    Params:
        gradients: List of gradient AtomicResult objects alternating between a
            "forward" and "backward" computation. NOTE: The last computation on the
            list is a basic energy calculation of the original geometry.
        dh: The displacement used for finite difference displacements of gradient
            geometries

    Note:
        Another way I've tested this algorithm is to compute the hessian using psi4
        and then using this algorithm at the save level of theory and then compare
        their eigenvalues. The results have always matched up to two decimal places.
        The matrices can't be compared directly because it appears psi4 does some sort
        of rotation on their matrix, so the eigenvalues are a better mechanism for
        comparison.
    """
    # Pop energy calculation of original geometry from gradients (last value in
    # gradients list)
    energy_output = gradients.pop()

    # # Verify data integrity of gradients
    # for gradient in gradients:

    dim = len(gradients[0].input_data.molecule.symbols) * 3
    hessian = np.zeros((dim, dim), dtype=float)

    for i, (forward, backward) in enumerate(zip_longest(*[iter(gradients)] * 2)):
        val = (forward.results.gradient - backward.results.gradient) / (dh * 2)  # type: ignore # noqa: E501
        hessian[i] = val.flatten()

    output = energy_output.model_dump()
    output["input_data"]["calctype"] = CalcType.hessian
    output["results"]["hessian"] = hessian

    return ProgramOutput[ProgramInput, SinglePointResults](**output)


@bigchem.task
def frequency_analysis(
    sp_output: ProgramOutput[ProgramInput, SinglePointResults], **kwargs
) -> ProgramOutput[ProgramInput, SinglePointResults]:
    """Adds geomeTRIC's frequency analysis results to hessian SinglePointOutput

    Params:
        sp_output: SinglePointOutput with .results.hessian value
        kwargs: Keywords passed to geomeTRIC's frequency_analysis function
            temperature: float - Temperature passed to the harmonic free energy module;
                default: 300.0
            pressure: float - Pressure passed to the harmonic free energy module;
                default: 1.0

    Returns:
        SinglePointOutput with additional results:
            freqs_wavenumber: List of vibrational frequencies in wavenumbers
            normal_modes_cartesian: List of normal modes in cartesian coordinates
            gibbs_free_energy: Gibbs free energy in Hartree

    """
    # Import here so client applications don't need to install geomeTRIC
    from geometric.normal_modes import frequency_analysis as geometric_freqs_analysis

    freqs, n_modes, g_tot = geometric_freqs_analysis(
        sp_output.input_data.molecule.geometry.flatten(),  # numpy array
        sp_output.results.hessian,  # type: ignore
        elem=sp_output.input_data.molecule.symbols,  # regular python list
        # Electronic energy passed to free energy module
        energy=sp_output.results.energy,  # type: ignore
        **kwargs,
    )
    output = sp_output.model_dump()
    output["results"].update(
        {
            "freqs_wavenumber": freqs.tolist(),
            "normal_modes_cartesian": n_modes,
            "gibbs_free_energy": g_tot,
        }
    )
    return ProgramOutput[ProgramInput, SinglePointResults](**output)


@bigchem.task
def add(x, y):
    """Add two numbers

    NOTE: Used for design testing
    """
    return x + y


@bigchem.task
def task_sum(values: List[Union[float, int]], extra: int = 0) -> Union[float, int]:
    """Sum all the values in a list

    NOTE: Used for design testing as a summation at the end of add (a chord)
        e.g., async_result = group(add.s(i, i) for i in range(max)) | csum.s()
    """
    values.append(extra)
    return sum(values)



"""Helper functions not for end users"""

from typing import List

import numpy as np
from qcio import CalcType, ProgramInput

from .config import settings


def _gradient_inputs(
    prog_input: ProgramInput, dh: float = settings.bigchem_default_hessian_dh
) -> List[ProgramInput]:
    """Create ProgramInput gradient calculations for a numerical hessian

    Params:
        prog_input: ProgramInput with keywords specific to the gradient computations
            that will comprise the hessian
        dh: Displacement for finite difference

    Returns:
        Flat list of ProgramInput gradient calculations with dh offset for each geometry
            value. The first ProgramInput represents a "forward" step by dh and the next
            ProgramInput represents a "backward" step by dh and so on.
    """
    as_dict = prog_input.model_dump()
    as_dict["calctype"] = CalcType.gradient
    grad_input = ProgramInput(**as_dict)

    gradients = []
    geometry = np.array(prog_input.molecule.geometry)
    # Get all indices in the 2D array as a list of pairs
    indices = np.indices(geometry.shape).reshape(2, -1).T

    for index in indices:
        # Need two new objects
        forward, backward = grad_input.model_copy(deep=True), grad_input.model_copy(
            deep=True
        )

        forward.molecule.geometry[tuple(index)] += dh
        backward.molecule.geometry[tuple(index)] -= dh

        gradients.append(forward)
        gradients.append(backward)

    return gradients



import ssl
from datetime import timedelta

from celery import Celery

from .config import settings

bigchem = Celery(
    # Name of top-level module is first argument
    # https://docs.celeryq.dev/en/stable/getting-started/first-steps-with-celery.html#application
    "bigchem",
    broker=settings.bigchem_broker_url,
    backend=settings.bigchem_backend_url,
)

bigchem.conf.update(
    # All configuration documentation here:
    # https://docs.celeryq.dev/en/stable/userguide/configuration.html
    # NOTE: Using pickle serializer so that chords receive python objects.
    # Can use JSON serializer json_dumps from qcio.utils if JSON is preferred.
    broker_connection_retry_on_startup=True,
    task_serializer="pickle",
    accept_content=["pickle"],
    result_serializer="pickle",
    task_track_started=True,
    task_acks_late=True,
    worker_prefetch_multiplier=settings.bigchem_prefetch_multiplier,
    worker_concurrency=settings.bigchem_worker_concurrency,
    result_expires=timedelta(seconds=settings.bigchem_result_expires),
)

# NOTE: If using SSL secured connection to broker, by default I am disabling
# client-side certificate verification. This makes things easier when running the
# broker behind a reverse proxy (like traefik) that dynamically provisions certificates.
# https://docs.celeryq.dev/en/stable/userguide/configuration.html#std-setting-broker_use_ssl
if "amqps" in settings.bigchem_broker_url:
    bigchem.conf.update(
        broker_use_ssl={
            "cert_reqs": ssl.CERT_NONE,
        },
    )



# https://github.com/python-poetry/poetry/pull/2366#issuecomment-652418094
from importlib import metadata

__version__ = metadata.version(__name__)

from .algos import *
from .canvas import *
from .tasks import *
